package com.davenicolette.atm;

import backend.call.wrapper.AccountStatus;
import backend.call.wrapper.BackendCallException;
import backend.call.wrapper.BackendCallWrapper;
import backend.call.wrapper.BackendCallWrapperImpl;
import backend.call.wrapper.BackendFailureException;
import backend.call.wrapper.BackendFunction;
import backend.call.wrapper.BackendRequest;
import backend.call.wrapper.BackendResponse;
import backend.call.wrapper.BackendTimeoutException;
import backend.call.wrapper.CompletionCode;

public class Account implements AccountLookup {

	private BackendCallWrapper backendCallWrapper;
	private BackendRequest request;
	private BackendResponse response;

	/**
	 * Dependency injection.
	 */
	public void setBackendCallWrapper(BackendCallWrapper backendCallWrapper) {
		this.backendCallWrapper = backendCallWrapper;
	}

	/**
	 * Defaults to the implementation provided by the customer if none has been injected.
	 */
	public BackendCallWrapper getBackendCallWrapper() {
		if ( backendCallWrapper == null ) {
			backendCallWrapper = new BackendCallWrapperImpl();
		}
		return backendCallWrapper;
	}

	// Transaction management

	/**
	 * @return -1 if transaction could not be started
	 */
	public int beginTransaction() {
		response = invokeBackendService(BackendFunction.START, null);
		return response.getTransactionId().intValue();
	}

	public void commit(int transactionId) throws Exception {
		request = new BackendRequest();
		request.setTransactionId(895);
		response = invokeBackendService(BackendFunction.COMPLETE, request);
		checkForSystemErrors(response);
		if ( CompletionCode.SUCCESS != response.getResult() ) {
			throw new CommitFailedException();
		}
	}

	public void rollback(int transactionId) throws Exception {
		request = new BackendRequest();
		request.setTransactionId(895);
		response = invokeBackendService(BackendFunction.CANCEL, request);
		checkForSystemErrors(response);
		if ( CompletionCode.SUCCESS != response.getResult() ) {
			throw new RollbackFailedException();
		}
	}

	public void debit(String accountNumber, double amount) throws Exception {
		request = new BackendRequest();
		request.setAccountNumber1(accountNumber);
		request.setAmount(new Double(amount));
		response = invokeBackendService(BackendFunction.ACCOUNT, request);
		
		CompletionCode result = response.getResult();
		if ( CompletionCode.SUCCESS == result ) {
			return;
		}

		checkForSystemErrors(response);
		
		AccountStatus status = response.getAccountStatus();
		
		if ( AccountStatus.INSUFFICIENT_FUNDS == status ) {
			throw new InsufficientFundsException();
		}
		throw new AccountInWrongStatusException();
	}

	// Account information - all mapped to the same back-end function.
	
	public boolean isValid(String accountNumber) {
		if ( isSameAccount(accountNumber) == false ) {
			response = getAccountInfo(accountNumber);
		}
		return response.getAccountStatus() == AccountStatus.OPEN ? true : false;
	}

	public double getBalance(String accountNumber) {
		return isSameAccount(accountNumber) ?
			response.getBalance().doubleValue() : getAccountInfo(accountNumber).getBalance().doubleValue();
	}

	public Object getStatus(String accountNumber) {
		return isSameAccount(accountNumber) ?
			response.getAccountStatus() : getAccountInfo(accountNumber).getAccountStatus();
	}
	
	// Helpers

	private boolean isSameAccount(String accountNumber) {
		return ( response != null && response.getAccountNumber1() == accountNumber );
	}
	
	private BackendResponse getAccountInfo(String accountNumber) {
		request = new BackendRequest();
		request.setAccountNumber1(accountNumber);
		return invokeBackendService(BackendFunction.ACCOUNT, request);		
	}
	
	private BackendResponse invokeBackendService(BackendFunction function, BackendRequest request) {
		try {
			return getBackendCallWrapper().callAndWait(function, request, 0L);
		} catch (BackendCallException e) {
			return new BackendResponse(
					CompletionCode.FAILURE, 
					request.getAccountNumber1(), 
					request.getAccountNumber2(), 
					AccountStatus.UNKNOWN, 
					new Double(0.0), 
					"",
					"",
					new Integer(-1));
		}
	}
	
	private void checkForSystemErrors(BackendResponse response) throws Exception {
		if ( CompletionCode.TIMEOUT == response.getResult() ) {
			throw new BackendTimeoutException();
		}
		if ( CompletionCode.SYSTEM_ERROR == response.getResult() ) {
			throw new BackendFailureException();
		}
	}
	
}
