package com.davenicolette.atm;

import static org.junit.Assert.*;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import org.junit.Before;
import org.junit.Test;

import backend.call.wrapper.AccountStatus;
import backend.call.wrapper.BackendCallException;
import backend.call.wrapper.BackendFailureException;
import backend.call.wrapper.BackendTimeoutException;
import backend.call.wrapper.BackendCallWrapper;
import backend.call.wrapper.BackendCallbackListener;
import backend.call.wrapper.BackendFunction;
import backend.call.wrapper.BackendRequest;
import backend.call.wrapper.BackendResponse;
import backend.call.wrapper.CompletionCode;

public class AccountLookupTest {
	private Account account;
	private FakeBackendCallWrapper backendCallWrapper = new FakeBackendCallWrapper();
	
	@Before
	public void initializeFixtures() {
		account = new Account();
	}
	
	@Test
	public void getBalanceRequestWithGoodResponse() {
		backendCallWrapper.setBackendResponse(new BackendResponse(
				CompletionCode.SUCCESS, "12345", "", AccountStatus.OPEN, new Double(2500.0), "", "", new Integer(-1)));
		account.setBackendCallWrapper(backendCallWrapper);
				
		assertEquals(2500.0, account.getBalance("12345"));
	}
	
	@Test
	public void isValidRequestWithGoodResponse() {
		backendCallWrapper.setBackendResponse(new BackendResponse(
				CompletionCode.SUCCESS, "12345", "", AccountStatus.OPEN, new Double(2500.0), "", "", new Integer(-1)));
		account.setBackendCallWrapper(backendCallWrapper);
				
		assertTrue("Account should be valid. ", account.isValid("12345"));
		assertEquals(AccountStatus.OPEN, account.getStatus("12345"));
	}
	
	@Test
	public void isValidRequestForNonexistentAccount() {
		backendCallWrapper.setBackendResponse(new BackendResponse(
				CompletionCode.SUCCESS, "12345", "", AccountStatus.UNKNOWN, new Double(2500.0), "", "", new Integer(-1)));
		account.setBackendCallWrapper(backendCallWrapper);
				
		assertFalse("Account should not be valid. ", account.isValid("12345"));
	}
	
	@Test
	public void successfulDebit() throws Exception {
		backendCallWrapper.setBackendResponse(new BackendResponse(
				CompletionCode.SUCCESS, "12345", "", AccountStatus.OPEN, new Double(2250.0), "", "", new Integer(-1)));
		account.setBackendCallWrapper(backendCallWrapper);
		
		account.debit("12345", 250.0);
		
		assertEquals(2250.0, account.getBalance("12345"));
	}
	
	@Test(expected=InsufficientFundsException.class)
	public void insufficientFundsForDebit() throws Exception {
		backendCallWrapper.setBackendResponse(new BackendResponse(
				CompletionCode.FAILURE, "12345", "", AccountStatus.INSUFFICIENT_FUNDS, new Double(50.0), "", "", new Integer(-1)));
		account.setBackendCallWrapper(backendCallWrapper);
		
		account.debit("12345", 250.0);
	}
	
	@Test(expected=AccountInWrongStatusException.class)
	public void problemWithAccountForDebit() throws Exception {
		backendCallWrapper.setBackendResponse(new BackendResponse(
				CompletionCode.FAILURE, "12345", "", AccountStatus.HOLD, new Double(0.0), "", "", new Integer(-1)));
		account.setBackendCallWrapper(backendCallWrapper);
		
		account.debit("12345", 250.0);
	}
	
	@Test(expected=BackendTimeoutException.class)
	public void requestTimedOutDuringDebitTransaction() throws Exception {
		backendCallWrapper.setBackendResponse(new BackendResponse(
				CompletionCode.TIMEOUT, "12345", "", AccountStatus.UNKNOWN, new Double(0.0), "", "", new Integer(-1)));
		account.setBackendCallWrapper(backendCallWrapper);
		
		account.debit("12345", 250.0);
	}
	
	@Test(expected=BackendFailureException.class)
	public void backendErrorDuringDebitTransaction() throws Exception {
		backendCallWrapper.setBackendResponse(new BackendResponse(
				CompletionCode.SYSTEM_ERROR, "12345", "", AccountStatus.UNKNOWN, new Double(0.0), "", "", new Integer(-1)));
		account.setBackendCallWrapper(backendCallWrapper);
		
		account.debit("12345", 250.0);
	}
	
	@Test
	public void beginTransactionSucceeds() {
		backendCallWrapper.setBackendResponse(new BackendResponse(
				CompletionCode.SUCCESS, "", "", AccountStatus.UNKNOWN, new Double(0.0), "", "", new Integer(895)));
		account.setBackendCallWrapper(backendCallWrapper);
		
		assertEquals(895, account.beginTransaction());
	}
	
	@Test
	public void beginTransactionFails() {
		backendCallWrapper.setBackendResponse(new BackendResponse(
				CompletionCode.FAILURE, "", "", AccountStatus.UNKNOWN, new Double(0.0), "", "", new Integer(-1)));
		account.setBackendCallWrapper(backendCallWrapper);
		
		assertEquals(-1, account.beginTransaction());
	}
	
	@Test
	public void commitSucceeds() throws Exception {
		backendCallWrapper.setBackendResponse(new BackendResponse(
				CompletionCode.SUCCESS, "", "", AccountStatus.UNKNOWN, new Double(0.0), "", "", new Integer(-1)));
		account.setBackendCallWrapper(backendCallWrapper);
		
		// No Exception == success
		account.commit(895);
	}
	
	@Test(expected=CommitFailedException.class)
	public void commitFails() throws Exception {
		backendCallWrapper.setBackendResponse(new BackendResponse(
				CompletionCode.FAILURE, "", "", AccountStatus.UNKNOWN, new Double(0.0), "", "", new Integer(-1)));
		account.setBackendCallWrapper(backendCallWrapper);
		
		// No exception == failure
		account.commit(895);
	}
	
	@Test
	public void rollbackSucceeds() throws Exception {
		backendCallWrapper.setBackendResponse(new BackendResponse(
				CompletionCode.SUCCESS, "", "", AccountStatus.UNKNOWN, new Double(0.0), "", "", new Integer(-1)));
		account.setBackendCallWrapper(backendCallWrapper);
		
		// No Exception == success
		account.rollback(895);
	}
	
	@Test(expected=RollbackFailedException.class)
	public void rollbackFails() throws Exception {
		backendCallWrapper.setBackendResponse(new BackendResponse(
				CompletionCode.FAILURE, "", "", AccountStatus.UNKNOWN, new Double(0.0), "", "", new Integer(-1)));
		account.setBackendCallWrapper(backendCallWrapper);
		
		// No exception == failure
		account.rollback(895);
	}
	
	class FakeBackendCallWrapper implements BackendCallWrapper {
		
		private BackendResponse response;
		public void setBackendResponse(BackendResponse response) {
			this.response = response;
		}
		
		public BackendResponse callAndWait(BackendFunction function,
				BackendRequest payload, double timeoutIntervalInMillis)
				throws BackendCallException {
			return response;
		}

		public void callAndForget(BackendFunction function,
				BackendRequest payload, BackendCallbackListener listener)
				throws BackendCallException {
		}
		
	}
}
