package com.davenicolette.atm;

import static org.junit.Assert.assertEquals;

import org.jmock.Expectations;
import org.jmock.Mockery;
import org.jmock.integration.junit4.JUnit4Mockery;
import org.junit.Before;
import org.junit.Test;

public class AtmTest {
	
	final Mockery mockery = new JUnit4Mockery();
	
	final DisplayManager displayManager = mockery.mock(DisplayManager.class);
	final AccountLookup account = mockery.mock(AccountLookup.class);
	final Dispenser dispenser = mockery.mock(Dispenser.class);
	final Logger logger = mockery.mock(Logger.class);

	private Atm atm;
	
	@Before
	public void setUpTest() {
		atm = new Atm();
		atm.setDisplayManager(displayManager);
		atm.setAccountLookup(account);
		atm.setDispenser(dispenser);
		atm.setLogger(logger);
	}

	@Test
	public void atmDisplaysWelcomeMessageWhenItIsReadyForInput() throws Exception {
		
		// Given
		Atm atm = new Atm();
		
		mockery.checking(new Expectations() {
			{ 
				one(displayManager).showMessage(with(AtmStatus.READY_FOR_CUSTOMER), with(0L));
			}
		});
		atm.setDisplayManager(displayManager);
		
		// When 
		atm.setStatus(AtmStatus.READY_FOR_CUSTOMER, 0L);
		
		// Then
		assertEquals(AtmStatus.READY_FOR_CUSTOMER, atm.getStatus());
		mockery.assertIsSatisfied();
		
	}
	
	@Test
	public void atmDisplaysOutOfServiceMessageWhenItIsNotReadyForInput() throws Exception {
		
		mockery.checking(new Expectations() {
			{ 
				one(displayManager).showMessage(with(AtmStatus.OUT_OF_SERVICE), with(0L));
			}
		});
		atm.setDisplayManager(displayManager);
		
		// When
		atm.setStatus(AtmStatus.OUT_OF_SERVICE, 0L);
		
		// Then
		mockery.assertIsSatisfied();
		
	}
	
	@Test
	public void cashWithdrawalNoErrors() throws Exception {
		
		// Set up the test conditions
		
		mockery.checking(new Expectations() {
			{
				one(account).beginTransaction(); will(returnValue(1));
				one(dispenser).requestCash(250.0);
				one(account).debit(with(aNonNull(String.class)), with(250.0));
				one(dispenser).dispenseCash();
				one(account).commit(1);
				one(logger).log(with(aNonNull(Object.class)));
				one(displayManager).showMessage(with(AtmStatus.NEW_TRANSACTION), with(0L));
			}
		});
		
		// Execute the code under test
		atm.withdrawCash("1234567890", 250.0);
		
		// Assert expected results
		mockery.assertIsSatisfied();
	}
	
	@Test
	public void atmLowCash() throws Exception {
		
		// Set up the test conditions

		mockery.checking(new Expectations() {
			{
				one(account).beginTransaction(); will(returnValue(1));
				one(dispenser).requestCash(250.0); will(throwException(new AtmLowCashException())); 
				one(dispenser).reset();
				one(logger).log(with(aNonNull(Object.class)));
				one(displayManager).showMessage(with(AtmStatus.ATM_LOW_CASH), with(0L));
				one(displayManager).showMessage(with(AtmStatus.OUT_OF_SERVICE), with(25000L));
			}
		});
		
		// Execute the code under test
		atm.withdrawCash("1234567890", 250.0);
		
		// Assert expected results
		mockery.assertIsSatisfied();
	}
	
	@Test
	public void insufficientFundsInAccount() throws Exception {
		
		// Set up the test conditions

		mockery.checking(new Expectations() {
			{
				one(account).beginTransaction(); will(returnValue(1));
				one(dispenser).requestCash(250.0); 
				one(account).debit(with(aNonNull(String.class)), with(250.0)); 
					will(throwException(new InsufficientFundsException()));
				one(account).rollback(1);
				one(dispenser).reset();
				one(logger).log(with(aNonNull(Object.class)));
				one(displayManager).showMessage(with(AtmStatus.INSUFFICIENT_FUNDS), with(0L));
				one(displayManager).showMessage(with(AtmStatus.NEW_TRANSACTION), with(25000L));
			}
		});
		
		// Execute the code under test
		atm.withdrawCash("1234567890", 250.0);
		
		// Assert expected results
		mockery.assertIsSatisfied();
	}
	
}
