package com.davenicolette.atm;

import static org.junit.Assert.*;
import static org.junit.Assert.assertEquals;

import org.jmock.Expectations;
import org.jmock.Mockery;
import org.jmock.integration.junit4.JUnit4Mockery;
import org.junit.Before;
import org.junit.Test;

import discrete.io.DioHandler;


public class DispenserTest {
	
	final Mockery mockery = new JUnit4Mockery();	
	final Logger logger = mockery.mock(Logger.class);
	final DioHandler dio = mockery.mock(DioHandler.class);
	
	FakeDispenser dispenser;
	
	@Before
	public void initializeFixtures() {
		dispenser = new FakeDispenser();
	}
	@Test
	public void dispenserKnowsHowMuchCashItHas() {
		dispenser.setCashAvailable(52168.0);
		assertEquals(52168.0, dispenser.cashAvailable());
	}
	
	@Test
	public void requestCashOperationSucceeds() throws Exception {
		dispenser.setCashAvailable(18000.0);
		dispenser.requestCash(400.0);
		assertEquals(400, dispenser.pendingAmountAfterRequestCash());
		assertEquals(18000, dispenser.cashAvailable());
	}
	
	@Test(expected=AtmLowCashException.class)
	public void requestCashInsufficientCashAvailable() throws Exception {
		dispenser.setCashAvailable(200.0);
		dispenser.requestCash(201.0);
	}
	
	@Test
	public void dispenseCashHappyPath() throws Exception {
		
		mockery.checking(new Expectations() {
			{ 
				exactly(1).of (dio).put(with(DioHandler.TRAY5)); // $100
				exactly(1).of (dio).get(with(DioHandler.FAULT), with(5L)); will(returnValue(0));
				exactly(1).of (dio).put(with(DioHandler.TRAY5)); // $100
				exactly(1).of (dio).get(with(DioHandler.FAULT), with(5L)); will(returnValue(0));
				exactly(1).of (dio).put(with(DioHandler.TRAY4)); // 1 $50
				exactly(1).of (dio).get(with(DioHandler.FAULT), with(5L)); will(returnValue(0));
				exactly(1).of (dio).put(with(DioHandler.TRAY3)); // 1 $20
				exactly(1).of (dio).get(with(DioHandler.FAULT), with(5L)); will(returnValue(0));
				exactly(1).of (dio).put(with(DioHandler.TRAY1)); // 1 $5
				exactly(1).of (dio).get(with(DioHandler.FAULT), with(5L)); will(returnValue(0));
			}
		});
		dispenser.setDioHandler(dio);
		dispenser.setCashAvailable(18000.0);

		dispenser.requestCash(275.0);
		dispenser.dispenseCash();

		assertEquals(17725.0, dispenser.cashAvailable());
		mockery.assertIsSatisfied();
	}
	
	@Test(expected=DeviceFaultException.class)
	public void dispenseCashWithDeviceFault() throws Exception {
		
		mockery.checking(new Expectations() {
			{ 
				exactly(1).of (dio).put(with(DioHandler.TRAY5)); // $100
				exactly(1).of (dio).get(with(DioHandler.FAULT), with(5L)); will(returnValue(0));
				exactly(1).of (dio).put(with(DioHandler.TRAY5)); // $100
				exactly(1).of (dio).get(with(DioHandler.FAULT), with(5L)); will(returnValue(0));
				exactly(1).of (dio).put(with(DioHandler.TRAY4)); // 1 $50
				exactly(1).of (dio).get(with(DioHandler.FAULT), with(5L)); will(returnValue(1));
			}
		});
		dispenser.setDioHandler(dio);
		dispenser.setCashAvailable(18000.0);

		dispenser.requestCash(275.0);
		dispenser.dispenseCash();
		
		assertEquals(17800.0, dispenser.cashAvailable());
		mockery.assertIsSatisfied();		
	}
	
	@Test(expected=InvalidOperationException.class)
	public void dispenseCashWithoutPriorRequest() throws Exception {
		
		mockery.checking(new Expectations() {
			{ 
			}
		});
		dispenser.setDioHandler(dio);
		dispenser.setCashAvailable(18000.0);

		dispenser.dispenseCash();
		
		assertEquals(18000.0, dispenser.cashAvailable());
		mockery.assertIsSatisfied();		
	}
	
	@Test(expected=InvalidOperationException.class)
	public void resetRequestedAmountToZero() throws Exception {
		dispenser.setCashAvailable(18000.0);
		dispenser.requestCash(500.0);
		dispenser.reset();
		dispenser.dispenseCash();
	}
	
	class FakeDispenser extends AnsiStandardDispenser {
		public void setCashAvailable(double amount) {
			cashAvailable = amount;
		}
		public double pendingAmountAfterRequestCash() {
			return pendingAmount;
		}
	}

}
